#!/usr/bin/env python
#
# Copyright (c) 2011 Clustrix, Inc.
#
# A script for monitoring a Clustrix database instance to be
# used within a Nagios system.

import MySQLdb
from optparse import OptionParser
import os
import sys

NAGIOS_OK   = 0
NAGIOS_WARN = 1
NAGIOS_CRIT = 2
NAGIOS_UNKN = 3
NAGIOS_STATUSES = ["OK", "WARN", "CRITICAL", "UNKNOWN"]

# we don't want exit handlers or exceptions (SystemExit) thrown
def exit(val):
    sys.stdout.flush()
    os._exit(val)

class NagiosCheck:
    checks = [ ]
    def __init__(self, f):
        self.check = f
        NagiosCheck.checks.append(self)

    def __call__(self):
        res = self.f(self.db)
        print res

    def execute(self, db, **kwargs):
        return self.check(db, **kwargs)

    def get_name(self):
        return self.check.func_name

    def get_descr(self):
        return self.check.func_doc

def simple_value_check(component, val, warn=None, crit=None):
    (ret,msg) = status_logic(val, warn, crit)
    print "%s: %s=%f | %s=%f" % (msg, component, val, component, val)
    return ret

def status_logic(val, warn=None, crit=None):
    """Evaluate the supplied value against the supplied warn and
    crit levels and return both the resulting status and the associated
    printable message"""

    ret = NAGIOS_OK
    if warn and float(val) >= float(warn):
        ret = NAGIOS_WARN
    if crit and float(val) >= float(crit):
        ret = NAGIOS_CRIT
    return(ret,NAGIOS_STATUSES[ret])

@NagiosCheck
def check_cpu(db, warn=80, crit=90, **kwargs):
    """
    Check that the cluster cpu utilization is within bounds.
    """
    cur = db.cursor()
    cur.execute("select round(avg(load*100)) from cpu_load where cpu<>0")
    return simple_value_check("CPU",
                              cur.fetchone()[0], warn=warn, crit=crit)

@NagiosCheck
def check_executing_sessions(db, warn=80, crit=90, **kwargs):
    """
    Check that the number of executing statements which have been
    running for more than 5 seconds is within a reasonable bound
    """
    cur = db.cursor()
    cur.execute("""select count(1)
                     from sessions
                    where statement_state = 'executing'
                      and time_in_state > 5""")
    return simple_value_check("EXEC_SESS",
                              cur.fetchone()[0], warn=warn, crit=crit)

@NagiosCheck
def check_long_running_queries(db, warn=1, **kwargs):
    """
    Check for queries which have been executing over 4 hours. 
    """
    cur = db.cursor()
    cur.execute("""select count(1) from sessions
                    where statement_state = 'executing'
                      and time_in_state > (60 * 60 * 4)""")
    return simple_value_check("LONG_RUNNING_QUERIES",
                              cur.fetchone()[0], warn=warn)

@NagiosCheck
def check_offline_slices(db, crit=1, **kwargs):
    """
    Check for number of offline slices. Any value over 0 indicates that
    some data is unavailable in the current membership. User quries will
    get errors.
    """
    cur = db.cursor()
    cur.execute("""select count(1) from (

                   select count(1)
                     from slices
                left join replicas using (slice)
                    where replicas.status = 1
                 group by slice
                   having count(1) < 1) unavailable""")
    return simple_value_check("OFFLINE_SLICES",
                              cur.fetchone()[0], crit=crit)

@NagiosCheck
def check_underprotected_slices(db, warn=1, **kwargs):
    """
    Check for any slices which have less than 2 replicas.
    """
    cur = db.cursor()
    cur.execute("""select count(1) from (
                   select count(1)
                     from slices
                left join replicas using (slice)
                    where replicas.status = 1
                 group by slice
                   having count(1) < 2) underprotected""")
    return simple_value_check("UNDERPROTECTED_SLICES",
                              cur.fetchone()[0], warn=warn)

@NagiosCheck
def fetch_stats(db, check_args=None, crit = None, warn=None, **kwargs):
    """
    Fetch current value from clustrix_statd tables.
    """
    if check_args is None:
        print "ERROR: you must specify a stat name"
        exit(NAGIOS_UNKN)
    stats = check_args
    cur = db.cursor()
    numrows = cur.execute("""SELECT name, value
                             FROM clustrix_statd.statd_current
                             JOIN clustrix_statd.statd_metadata USING (id)
                             WHERE name IN ('%s')""" % "','".join(stats))
    if not numrows:
        print 'UNK: no matching rows found in statd tables'
        exit(NAGIOS_UNKN)
    name_vals = []
    name_vals_full = []
    # loop through all statd rows returned
    for r in cur.fetchall():
        (name,value) = r
        # choose integer/float format as appropriate
        if int(value) == value:
            fmt = "%s=%d"
        else:
            fmt = "%s=%f"
        # full format prepends the Nagios status ("OK", "CRITICAL", etc)
        fmt_full = "%s - " + fmt
        (status,msg) = status_logic(value, warn, crit)
        name_vals.append(fmt % (name, value))
        name_vals_full.append(fmt_full % (msg, name, value))
    print "%s | %s" % (' '.join(name_vals_full), ' '.join(name_vals))
    return status

@NagiosCheck
def fetch_session_stats(db, **kwargs):
    """
    Print out 4 specific statd values. This is a deprecated function that is
    being kept around to enhance compatability. Note that the code is forked
    off of "fetch_stats" with hardcoded statd values used instead of
    commandline parameters. The four statd parameters used are:
    'clustrix.sessions'
    'clustrix.sessions.transactions.open.executing'
    'clustrix.sessions.transactions.open.executing.avg_time'
    'clustrix.sessions.transactions.open.executing.max_time'
    """
    # hardcode in statd parameter for function
    stats = ['clustrix.sessions',
             'clustrix.sessions.transactions.open.executing',
             'clustrix.sessions.transactions.open.executing.avg_secs',
             'clustrix.sessions.transactions.open.executing.max_secs']
    stat_descriptions = ['CONNECTIONS',
                         'EXECUTING_SESSIONS',
                         'AVG_QUERY_AGE',
                         'MAX_QUERY_AGE']

    # hardcode in a lack of crit / warn values for this deprecated function;
    # users should use fetch_stats to achieve this functionality
    crit = None
    warn = None

    cur = db.cursor()
    numrows = cur.execute("""SELECT name, value
                             FROM clustrix_statd.statd_current
                             JOIN clustrix_statd.statd_metadata USING (id)
                             WHERE name IN ('%s')""" % "','".join(stats))
    if not numrows:
        print 'UNK: no matching rows found in statd tables'
        exit(NAGIOS_UNKN)
    name_vals = []
    name_vals_full = []
    rows = cur.fetchall()

    for (stat, desc) in zip(stats, stat_descriptions):
        # loop through all statd rows returned
        for r in rows:
            (name,value) = r
            if name == stat:
                break
        else:
            print 'ERROR: unable to find desired values in statd'
            exit(NAGIOS_UNKN)

        # choose integer/float format as appropriate
        if int(value) == value:
            fmt = "%s=%d"
        else:
            fmt = "%s=%f"
        # full format prepends the Nagios status ("OK", "CRITICAL", etc)
        fmt_full = "%s - " + fmt
        (status,msg) = status_logic(value, warn, crit)
        name_vals.append(fmt % (desc, value))
        name_vals_full.append(fmt_full % (msg, desc, value))
    print "%s | %s" % (' '.join(name_vals_full), ' '.join(name_vals))
    return status
 
def list_checks(*args, **kwargs):
    print "Available checks:"
    for check in NagiosCheck.checks:
        print check.get_name()
        print check.get_descr()
        print
    exit(0)

usage = "Usage: %prog [options] <command>"
parser = OptionParser(usage=usage)
parser.add_option("-H", "--hostname", dest="hostname",
                  help="any hostname for a clustrix system")
parser.add_option("-P", "--port", dest="port",
                  help="port to connect to", default=3306)
parser.add_option("-u", "--user", dest="user",
                  help="database user to connect as",
                  default="root")
parser.add_option("-p", "--password", help="account password",
                  dest="password", default="")
parser.add_option("-L", "--list-checks", help="list available checks",
                  action="callback", callback=list_checks)
parser.add_option("-w", "--warn", help="warning threshold",
                  dest="warn", type="float")
parser.add_option("-c", "--crit", help="critical threshold",
                  dest="crit", type="float")

(opts, clargs) = parser.parse_args()

if len(clargs) < 1:
    parser.print_help()
    print "ERROR: you must specify a check name; use -L to list available checks"
    exit(NAGIOS_UNKN)

if not opts.hostname:
    print "ERROR: hostname required"
    exit(NAGIOS_UNKN)

check_name = clargs[0]

check = None
for c in NagiosCheck.checks:
    if check_name == c.get_name():
        check = c
        break

if not check:
    print "ERROR: check '%s' not found" % check_name
    exit(NAGIOS_UNKN)

try:
    db = MySQLdb.connect(host=opts.hostname, db="system",
                         port=opts.port, user=opts.user,
                         passwd=opts.password)
    args = { }
    if opts.warn != None:
        args['warn'] = opts.warn
    if opts.crit != None:
        args['crit'] = opts.crit
    if len(clargs) > 1:
        args['check_args'] = clargs[1:]
    result = check.execute(db, **args)
    exit(result)
except Exception, e:
    print e
    exit(NAGIOS_UNKN)
